==> file_storage_handler_abm-8.0.67643a0.tar.gz: file_storage_handler_abm-8.0.67643a0/file_storage_handler/controller.py <==
#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import logging
from dataclasses import dataclass
from typing import Type

from aiohttp import ClientResponseError
from dict_caster.extras import first
from file_storage.abstract_file_storage import AbstractFileStorage

from .database import Database, WrappedSession
from .entities.link_info import LinkInfo
from .entities.thumbnail_info import ThumbnailInfo
from .extra import silent, AnySession
from .file_downloader import FileDownloader
from .file_keeper import FileKeeper
from .link_keeper import LinkKeeper, LI
from .memory_zipper import MemoryZipper
from .thumbnail.thumbnail_provider import ThumbnailProvider
from .tools.exceptions import FileStorageFullException


logger = logging.getLogger(__name__)


class Controller:
    @dataclass
    class Config:
        pass

    @dataclass
    class Context:
        database: Database
        file_storage: AbstractFileStorage
        thumbnail_provider: ThumbnailProvider
        file_downloader: FileDownloader

    def __init__(self,
                 config: Config, context: Context,
                 link_class: Type[LI] = LinkInfo, thumbnail_class: Type[LI] = ThumbnailInfo,
                 ) -> None:
        self.config = config
        self.context = context

        self.file_keeper = FileKeeper(FileKeeper.Context(self.context.database, self.context.file_storage))
        self.link_keeper = LinkKeeper(
            LinkKeeper.Context(self.file_keeper, self.context.database), link_class=link_class
        )
        self.thumbnail_keeper = LinkKeeper(
            LinkKeeper.Context(self.file_keeper, self.context.database), link_class=thumbnail_class
        )
        self.zipper = MemoryZipper(MemoryZipper.Context(self.link_keeper, self.context.database))
        self._files_total_size: int | None = None

    def _publish(self):
        ...

    async def _find_thumbnail_key_by_file_key(self, file_info_key: str, session:  WrappedSession) -> str | None:
        if link_infos_with_same_key := await self.link_keeper.get_link_infos_by_file_info_key(file_info_key, session):
            links_ids = [link.id for link in link_infos_with_same_key]
            if thumbnail_infos := await self.thumbnail_keeper.get_link_infos(links_ids, session, limit=1):
                return thumbnail_infos[0].file_info_key

    async def _check_limit_file_storage(self, content: bytes, instance: str, session: WrappedSession) -> None:
        if self._files_total_size is None:
            self._files_total_size = await self.get_files_total_size(instance, session) or 0
        content_size = len(content)
        free_storage_size = self.context.file_storage.config.size_b - self._files_total_size \
            if self.context.file_storage.config.size_b > self._files_total_size else 0 # TODO: attr size_b will be added in library 'file-storage-abm'
        logger.info(f"Free storage size {free_storage_size} bytes")
        if content_size > free_storage_size:
            raise FileStorageFullException(f'Not enough disk space. Free storage size {free_storage_size} bytes, '
                                           f'but need {content_size} bytes')
        self._files_total_size += content_size

    async def upload_from(self, file_url: str, instance: str, session: AnySession | None = None) -> LI:
        try:
            file_metadata = await self.context.file_downloader.head(file_url)
        except ClientResponseError as er:
            logger.error(f"Received file_metadata failed: {repr(er)}")
            file_metadata = None
        file_info_key = file_metadata.cleared_etag if file_metadata else None
        if file_info_key and await self.file_keeper.head(file_info_key, nullable=True):
            return await self.add_link(
                file_info_key, file_metadata.filename, extension=file_metadata.extension, session=session
            )
        logger.info(f"file '{file_url}' will be downloaded and uploaded")
        filename, content = await self.context.file_downloader.get(file_url)
        return first(await self.upload(content, filename, instance, session=session))

    async def upload(
            self, content: bytes, filename: str, instance: str, new_link_id: str | None = None, session: AnySession = None,
            is_thumbnail_required: bool = False) -> tuple[LinkInfo, ThumbnailInfo | None]:
        async with self.context.database.ensure_session(session) as session:
            await self._check_limit_file_storage(content, instance, session)
            link = await self.link_keeper.upload(content, instance, filename, link_id=new_link_id, session=session)
            thumbnail = await self.add_thumbnail(link, content, session) if is_thumbnail_required else None
        return link, thumbnail

    async def add_thumbnail(self, link: LI, content: bytes, session: AnySession | None = None) -> ThumbnailInfo | None:
        async with self.context.database.ensure_session(session) as session:
            if thumbnail_key := await self._find_thumbnail_key_by_file_key(link.file_info_key, session):
                logger.info("Found existing thumbnail, using it")
                return await self.thumbnail_keeper.add(
                    thumbnail_key, link.filename, link.id, link.extension, session
                )
            elif thumbnail := await self.context.thumbnail_provider.provide(link.extension, content):
                logger.info("Existing thumbnail not found, created new")
                return await self.thumbnail_keeper.upload(thumbnail, link.filename, link.id, session)

    async def add_link(
            self, file_info_key: str, filename: str, link_id: str | None = None,
            extension: str | None = None, session: AnySession | None = None, link_kwargs: dict | None = None
    ) -> LI:
        return await self.link_keeper.add(file_info_key, filename, link_id, extension, session, link_kwargs)

    async def get(self, link_id: str, byte_range: range | None = None,
                  session: AnySession = None) -> tuple[LinkInfo, bytes]:
        link_info, file_content = await self.link_keeper.get(link_id, byte_range=byte_range, session=session)
        return link_info, file_content

    async def get_thumbnail(self, link_id: str, byte_range: range | None = None,
                            session: AnySession = None) -> tuple[LinkInfo, bytes]:
        link_info, file_content = await self.thumbnail_keeper.get(link_id, byte_range=byte_range, session=session)
        return link_info, file_content

    async def read(self, file_info_key: str, byte_range: range | None = None) -> bytes:
        return await self.link_keeper.read(file_info_key, byte_range)

    async def head(self, link_id: str, session: AnySession = None) -> LinkInfo:
        return await self.link_keeper.head(link_id, session=session)

    async def delete(self, link_id: str, session: AnySession = None) -> LinkInfo:
        async with self.context.database.ensure_session(session) as session:
            result = await self.link_keeper.delete(link_id, session)
            await silent(self.thumbnail_keeper.delete(link_id, session))
        return result

    async def zip_files(self, link_ids: list[str], session: AnySession = None) -> tuple[LinkInfo, bytes]:
        return await self.zipper.zip_files(link_ids, session=session)

    async def get_files_total_size(self, instance: str, session: AnySession = None) -> int | None:
        self._files_total_size = await self.file_keeper.get_files_total_size(instance, session)
        return self._files_total_size

==> file_storage_handler_abm-8.0.67655a0.tar.gz: file_storage_handler_abm-8.0.67655a0/file_storage_handler/controller.py <==
#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import logging
from dataclasses import dataclass
from typing import Type

from aiohttp import ClientResponseError
from dict_caster.extras import first
from file_storage.abstract_file_storage import AbstractFileStorage

from .database import Database, WrappedSession
from .entities.link_info import LinkInfo
from .entities.thumbnail_info import ThumbnailInfo
from .extra import silent, AnySession
from .file_downloader import FileDownloader
from .file_keeper import FileKeeper
from .link_keeper import LinkKeeper, LI
from .memory_zipper import MemoryZipper
from .thumbnail.thumbnail_provider import ThumbnailProvider
from .tools.exceptions import FileStorageFullException


logger = logging.getLogger(__name__)


class Controller:
    @dataclass
    class Config:
        pass

    @dataclass
    class Context:
        database: Database
        file_storage: AbstractFileStorage
        thumbnail_provider: ThumbnailProvider
        file_downloader: FileDownloader

    def __init__(self,
                 config: Config, context: Context,
                 link_class: Type[LI] = LinkInfo, thumbnail_class: Type[LI] = ThumbnailInfo,
                 ) -> None:
        self.config = config
        self.context = context

        self.file_keeper = FileKeeper(FileKeeper.Context(self.context.database, self.context.file_storage))
        self.link_keeper = LinkKeeper(
            LinkKeeper.Context(self.file_keeper, self.context.database), link_class=link_class
        )
        self.thumbnail_keeper = LinkKeeper(
            LinkKeeper.Context(self.file_keeper, self.context.database), link_class=thumbnail_class
        )
        self.zipper = MemoryZipper(MemoryZipper.Context(self.link_keeper, self.context.database))
        self.storage_total_bytes: int | None = None

    def _publish(self):
        ...

    async def _find_thumbnail_key_by_file_key(self, file_info_key: str, session:  WrappedSession) -> str | None:
        if link_infos_with_same_key := await self.link_keeper.get_link_infos_by_file_info_key(file_info_key, session):
            links_ids = [link.id for link in link_infos_with_same_key]
            if thumbnail_infos := await self.thumbnail_keeper.get_link_infos(links_ids, session, limit=1):
                return thumbnail_infos[0].file_info_key

    async def _check_file_storage_limit(self, content: bytes, instance: str, session: WrappedSession) -> None:
        if self.storage_total_bytes is None:
            self.storage_total_bytes = await self.get_storage_total_bytes(instance, session)
        content_size = len(content)
        free_storage_size = self.context.file_storage.config.size_b - self.storage_total_bytes \
            if self.context.file_storage.config.size_b > self.storage_total_bytes else 0 # TODO: attr size_b will be added in library 'file-storage-abm'
        logger.info(f"Free storage size {free_storage_size} bytes")
        if content_size > free_storage_size:
            raise FileStorageFullException(f'Not enough disk space. Free storage size {free_storage_size} bytes, '
                                           f'but need {content_size} bytes')
        self.storage_total_bytes += content_size

    async def upload_from(self, file_url: str, instance: str, session: AnySession | None = None) -> LI:
        try:
            file_metadata = await self.context.file_downloader.head(file_url)
        except ClientResponseError as er:
            logger.error(f"Received file_metadata failed: {repr(er)}")
            file_metadata = None
        file_info_key = file_metadata.cleared_etag if file_metadata else None
        if file_info_key and await self.file_keeper.head(file_info_key, nullable=True):
            return await self.add_link(
                file_info_key, file_metadata.filename, extension=file_metadata.extension, session=session
            )
        logger.info(f"file '{file_url}' will be downloaded and uploaded")
        filename, content = await self.context.file_downloader.get(file_url)
        return first(await self.upload(content, filename, instance, session=session))

    async def upload(
            self, content: bytes, filename: str, instance: str, new_link_id: str | None = None, session: AnySession = None,
            is_thumbnail_required: bool = False) -> tuple[LinkInfo, ThumbnailInfo | None]:
        async with self.context.database.ensure_session(session) as session:
            await self._check_file_storage_limit(content, instance, session)
            link = await self.link_keeper.upload(content, instance, filename, link_id=new_link_id, session=session)
            thumbnail = await self.add_thumbnail(link, content, session) if is_thumbnail_required else None
        return link, thumbnail

    async def add_thumbnail(self, link: LI, content: bytes, session: AnySession | None = None) -> ThumbnailInfo | None:
        async with self.context.database.ensure_session(session) as session:
            if thumbnail_key := await self._find_thumbnail_key_by_file_key(link.file_info_key, session):
                logger.info("Found existing thumbnail, using it")
                return await self.thumbnail_keeper.add(
                    thumbnail_key, link.filename, link.id, link.extension, session
                )
            elif thumbnail := await self.context.thumbnail_provider.provide(link.extension, content):
                logger.info("Existing thumbnail not found, created new")
                return await self.thumbnail_keeper.upload(thumbnail, link.filename, link.id, session)

    async def add_link(
            self, file_info_key: str, filename: str, link_id: str | None = None,
            extension: str | None = None, session: AnySession | None = None, link_kwargs: dict | None = None
    ) -> LI:
        return await self.link_keeper.add(file_info_key, filename, link_id, extension, session, link_kwargs)

    async def get(self, link_id: str, byte_range: range | None = None,
                  session: AnySession = None) -> tuple[LinkInfo, bytes]:
        link_info, file_content = await self.link_keeper.get(link_id, byte_range=byte_range, session=session)
        return link_info, file_content

    async def get_thumbnail(self, link_id: str, byte_range: range | None = None,
                            session: AnySession = None) -> tuple[LinkInfo, bytes]:
        link_info, file_content = await self.thumbnail_keeper.get(link_id, byte_range=byte_range, session=session)
        return link_info, file_content

    async def read(self, file_info_key: str, byte_range: range | None = None) -> bytes:
        return await self.link_keeper.read(file_info_key, byte_range)

    async def head(self, link_id: str, session: AnySession = None) -> LinkInfo:
        return await self.link_keeper.head(link_id, session=session)

    async def delete(self, link_id: str, session: AnySession = None) -> LinkInfo:
        async with self.context.database.ensure_session(session) as session:
            result = await self.link_keeper.delete(link_id, session)
            await silent(self.thumbnail_keeper.delete(link_id, session))
        return result

    async def zip_files(self, link_ids: list[str], session: AnySession = None) -> tuple[LinkInfo, bytes]:
        return await self.zipper.zip_files(link_ids, session=session)

    async def get_storage_total_bytes(self, instance: str, session: AnySession = None) -> int:
        return await self.file_keeper.get_files_total_size(instance, session) or 0
